import ast
import copy
import csv
import json
import os
import re
import sys
import time
import warnings
from typing import Dict

import anthropic
import numpy as np
import pandas as pd
import requests
import torch
import openai
from together import Together
from datasets import Dataset
from openai import OpenAI
from sklearn.model_selection import train_test_split
from tqdm import tqdm
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, AutoConfig, AutoModelForCausalLM

try:
    from vllm import LLM
    VLLM_AVAILABLE = True
except ImportError:
    VLLM_AVAILABLE = False
    print("vLLM not imported.")

def calculate_accuracy(predictions, ground_truth) -> float:
    """
    Calculate the accuracy percentage between predictions and ground truths.

    Args:
    predictions (list): A list of predicted values.
    ground_truth (list): A list of actual values.

    Returns:
    float: The accuracy percentage rounded to two decimal places.

    Raises:
    ValueError: If the input lists have different lengths.
    """
    if len(predictions) != len(ground_truth):
        print(f"Predictions count: {len(predictions)}")
        print(f"Ground truth count: {len(ground_truth)}")
        raise ValueError("Input lists must have the same length")

    correct_count = sum(1 for pred, truth in zip(predictions, ground_truth) if pred == truth)
    total_count = len(predictions)

    accuracy = round(correct_count * 100 / total_count, 2)
    return accuracy

def few_shot_context_relevance_scoring(
    system_prompt: str, query: str, document: str, gpt_model: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the relevance of a query given a document using few-shot examples.
    This function constructs a user prompt with few-shot examples (if provided) and the current query,
    document, and answer. It then queries an OpenAI model to determine if the query is context relevant to the
    document.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query to be evaluated.
    document (str): The document to be evaluated.
    gpt_model (str): The model identifier for OpenAI's GPT.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating context relevance (1 for relevant, 0 for not relevant).

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    for attempt in range(5):
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += f"Question: {query}\nDocument: {document}\nLabel: "

            time.sleep(request_delay)

            if debug_mode: 
                print("------------------------------------------")
                print("Context Relevance Evaluation:")
                print(user_prompt)

            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content

            if debug_mode: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"

            if re.search(yes, final_response): 
                if debug_mode:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0

def few_shot_answer_faithfulness_scoring(
    system_prompt: str, query: str, document: str, answer: str, gpt_model: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the faithfulness of an answer given a document and a query using few-shot examples.
    This function constructs a user prompt with few-shot examples (if provided) and the current query,
    document, and answer. It then queries an OpenAI model to determine if the answer is faithful.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query related to the document.
    document (str): The document to be evaluated.
    answer (str): The answer to be evaluated for faithfulness.
    gpt_model (str): The model identifier for OpenAI's GPT.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating answer faithfulness 1 if the answer is faithful, 0 otherwise.

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    for attempt in range(5): 
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    user_prompt += f"Answer: {few_shot_examples.iloc[row]['Answer']}\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += f"Question: {query}\nDocument: {document}\nAnswer: {answer}\nLabel: "

            time.sleep(request_delay) # Model query delay chosen by user
            
            if debug_mode: 
                print("------------------------------------------")
                print("Answer Faithfulness Evaluation")
                print(user_prompt)

            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content

            if debug_mode: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"

            if re.search(yes, final_response): 
                if debug_mode:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0

def few_shot_answer_relevance_scoring(
    system_prompt: str, query: str, document: str, answer: str, gpt_model: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the relevance of an answer given a document and a query using few-shot examples.
    This function constructs a user prompt with few-shot examples (if provided) and the current query,
    document, and answer. It then queries an OpenAI model to determine if the answer is faithful.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query related to the document.
    document (str): The document to be evaluated.
    answer (str): The answer to be evaluated for relevance.
    gpt_model (str): The model identifier for OpenAI's GPT.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating answer relevance (1 for relevant, 0 for not relevant).

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    for attempt in range(5): 
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    user_prompt += f"Answer: {few_shot_examples.iloc[row]['Answer']}\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += f"Question: {query}\nDocument: {document}\nAnswer: {answer}\nLabel: "

            time.sleep(request_delay)
            
            if debug_mode: 
                print("------------------------------------------")
                print("Answer Relevance Evaluation:")
                print(user_prompt)

            messages = [
                            {
                                "role": "system",
                                "content": system_prompt
                            },
                            {
                                "role": "user",
                                "content": user_prompt
                            }
                        ]
            response = openai.chat.completions.create(
                                model=gpt_model,
                                messages=messages
                            )    
            final_response = response.choices[0].message.content

            if debug_mode: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"

            if re.search(yes, final_response): 
                if debug_mode:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0

def few_shot_context_relevance_scoring_togetherai(
    system_prompt: str, query: str, document: str, model_choice: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the relevance of a query given a document using few-shot examples with a specified model from TogetherAI.
    This function constructs a user prompt with optional few-shot examples and the current query and document.
    It then queries a TogetherAI model to determine if the query is contextually relevant to the document.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query related to the document.
    document (str): The document to be evaluated.
    model_choice (str): The model identifier for TogetherAI's API.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating context relevance (1 for relevant, 0 for not relevant).

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")
    if not TOGETHER_API_KEY:
        sys.exit("TOGETHER_API_KEY environment variable not set. Please set the variable.")
    
    for attempt in range(5): 
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += f"Label: {current_label}\n\n"

            user_prompt += f"Question: {query}\nDocument: {document}\nLabel: "

            time.sleep(request_delay)

            if debug_mode: 
                print("Context Relevance Evaluation:")
                print(user_prompt)

            client = Together(api_key=TOGETHER_API_KEY)

            # Enable streaming to receive responses as they are generated, reducing wait times.
            chat_completion = client.chat.completions.create(
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt}
                ],
                model=model_choice,
                stream=True
            )

            responses = [chunk.choices[0].delta.content for chunk in chat_completion]
            final_response = " ".join(responses)

            if debug_mode: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"
            
            if re.search(yes, final_response): 
                if debug_mode:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0

def few_shot_answer_faithfulness_scoring_togetherai(
    system_prompt: str, query: str, document: str, answer: str, model_choice: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the faithfulness of an answer given a document using few-shot examples with a specified model from TogetherAI.
    This function constructs a user prompt with optional few-shot examples and the current query, document, and answer.
    It then queries a TogetherAI model to determine if the answer is faithful.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query to be evaluated. 
    document (str): The document to be evaluated.
    answer (str): The answer to be evaluated for faithfulness.
    model_choice: The model identifier for TogetherAI's API.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating answer faithfulness (1 for faithful, 0 for not faithful).

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")
    if not TOGETHER_API_KEY:
        sys.exit("TOGETHER_API_KEY environment variable not set. Please set the variable.")
    for attempt in range(5):
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    user_prompt += f"Answer: {few_shot_examples.iloc[row]['Answer']}\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            

            user_prompt += f"Question: {query}\nDocument: {document}\nAnswer: {answer}\nLabel: "

            time.sleep(request_delay)

            if debug_mode: 
                print("------------------------------------------")
                print("Answer Faithfulness Evaluation:")
                print(user_prompt)

            client = Together(api_key=TOGETHER_API_KEY)

            # Enable streaming to receive responses as they are generated, reducing wait times.
            chat_completion = client.chat.completions.create(
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt}
                ],
                model=model_choice,
                stream=True
            )

            responses = []
            for chunk in chat_completion:
                responses.append(chunk.choices[0].delta.content)

            final_response = " ".join(responses)

            if debug_mode: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"

            if re.search(yes, final_response): 
                if debug_mode:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0

def few_shot_answer_relevance_scoring_togetherai(
    system_prompt: str, query: str, document: str, answer: str, model_choice: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the relevance of an answer given a document using few-shot examples with a specified model from TogetherAI.
    This function constructs a user prompt with optional few-shot examples and the current query, document, and answer.
    It then queries a TogetherAI model to determine if the answer is relevant.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query to be evaluated. 
    document (str): The document to be evaluated.
    answer (str): The answer to be evaluated for relevance.
    model_choice: The model identifier for TogetherAI's API.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating answer relevance (1 for faithful, 0 for not faithful).

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")
    if not TOGETHER_API_KEY:
        sys.exit("TOGETHER_API_KEY environment variable not set. Please set the variable.")
    
    for attempt in range(5):
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    user_prompt += f"Answer: {few_shot_examples.iloc[row]['Answer']}\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += f"Question: {query}\nDocument: {document}\nAnswer: {answer}\nLabel: "

            time.sleep(request_delay)

            if debug_mode: 
                print("------------------------------------------")
                print("Answer Relevance Evaluation:")
                print(user_prompt)

            client = Together(api_key=TOGETHER_API_KEY)

            chat_completion = client.chat.completions.create(
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt},
                ],
                model=model_choice,
                stream=True
            )

            responses = []
            
            for chunk in chat_completion:
                responses.append(chunk.choices[0].delta.content)

            final_response = " ".join(responses)
            
            if debug_mode: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"
            
            if re.search(yes, final_response): 
                if debug_mode:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0
            
def few_shot_context_relevance_scoring_claude(
    system_prompt: str, query: str, document: str, model_choice: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the relevance of a query given a document using few-shot examples with a specified model from TogetherAI.
    This function constructs a user prompt with optional few-shot examples and the current query and document.
    It then queries an Anthropic model to determine if the query is contextually relevant to the document.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query related to the document.
    document (str): The document to be evaluated.
    model_choice (str): The model identifier for Anthropic's API.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating context relevance (1 for relevant, 0 for not relevant).

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    ANTHROPIC_API_KEY = os.getenv("ANTHROPIC_API_KEY")
    if not ANTHROPIC_API_KEY:
        sys.exit("ANTHROPIC_API_KEY environment variable not set. Please set the variable.")

    for attempt in range(5):
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"
            
            user_prompt += f"Question: {query}\nDocument: {document}\nLabel: "

            time.sleep(request_delay)

            if debug_mode: 
                print("------------------------------------------")
                print("Context Relevance Evaluation:")
                print(user_prompt)

            client = anthropic.Anthropic(
            api_key=ANTHROPIC_API_KEY,
            )

            responses = []

            with client.messages.stream( 
                max_tokens=1024, 
                system=system_prompt.strip(),
                messages=[{"role": "user", "content": user_prompt.strip()}], 
                model=model_choice
            ) as stream: 
                for text in stream.text_stream: 
                    responses.append(text)

            final_response = " ".join(responses)

            if debug_mode is True: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"

            if re.search(yes, final_response): 
                if debug_mode is True:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode is True:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0

def few_shot_answer_faithfulness_scoring_claude(
    system_prompt: str, query: str, document: str, answer: str, model_choice: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the faithfulness of an answer given a document and a query using few-shot examples.
    This function constructs a user prompt with few-shot examples (if provided) and the current query,
    document, and answer. It then queries an Anthropic model to determine if the answer is faithful.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query related to the document.
    document (str): The document to be evaluated.
    answer (str): The answer to be evaluated for faithfulness.
    model_choice (str): The model identifier for Anthropic's API.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating answer faithfulness 1 if the answer is faithful, 0 otherwise.

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    ANTHROPIC_API_KEY = os.getenv("ANTHROPIC_API_KEY")
    if not ANTHROPIC_API_KEY:
        sys.exit("ANTHROPIC_API_KEY environment variable not set. Please set the variable.")

    for attempt in range(5):
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    user_prompt += f"Answer: {few_shot_examples.iloc[row]['Answer']}\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"

            user_prompt += f"Question: {query}\nDocument: {document}\nAnswer: {answer}\nLabel: "

            time.sleep(request_delay)

            if debug_mode: 
                print("------------------------------------------")
                print("Answer Faithfulness Evaluation")
                print(user_prompt)

            client = anthropic.Anthropic(
            api_key=ANTHROPIC_API_KEY,
            )

            responses = []

            with client.messages.stream( 
                max_tokens=1024, 
                system=system_prompt.strip(),
                messages=[{"role": "user", "content": user_prompt.strip()}], 
                model=model_choice
            ) as stream: 
                for text in stream.text_stream: 
                    responses.append(text)

            final_response = " ".join(responses)

            if debug_mode: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"
            
            if re.search(yes, final_response): 
                if debug_mode is True:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode is True:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0

def few_shot_answer_relevance_scoring_claude(
    system_prompt: str, query: str, document: str, answer: str, model_choice: str, 
    query_id: str, debug_mode: bool, request_delay: int, 
    failed_extraction_count: Dict[str,int] = {'failed': 0}, 
    few_shot_examples=None
) -> int:
    """
    Evaluates the relevance of an answer given a document and a query using few-shot examples.
    This function constructs a user prompt with few-shot examples (if provided) and the current query,
    document, and answer. It then queries an Anthropic model to determine if the answer is relevant.

    Args:
    system_prompt (str): The prompt for the system role in the conversation.
    query (str): The query related to the document.
    document (str): The document to be evaluated.
    answer (str): The answer to be evaluated for relevance.
    model_choice (str): The model identifier for an Anthropic's API.
    query_id (str): The identifier for the query column in few_shot_examples.
    debug_mode (bool): Flag to turn on debug mode for additional output.
    request_delay (int): Time in seconds to delay the request (simulating network latency).
    failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
    few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

    Returns:
    int: The label indicating answer relevance 1 if the answer is relevant, 0 otherwise.

    Raises:
    Warning: If incorrect labels are detected in the few-shot examples.
    """
    ANTHROPIC_API_KEY = os.getenv("ANTHROPIC_API_KEY")
    if not ANTHROPIC_API_KEY:
        sys.exit("ANTHROPIC_API_KEY environment variable not set. Please set the variable.")

    for attempt in range(5):
        try:
            user_prompt = ""
            if few_shot_examples is not None:
                for row in range(len(few_shot_examples)):
                    user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                    user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                    user_prompt += f"Answer: {few_shot_examples.iloc[row]['Answer']}\n"
                    current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
                    if current_label in {1, 1.0}:
                        warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                        current_label = "[[Yes]]"
                    elif current_label in {0, 0.0}:
                        warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                        current_label = "[[No]]"
                    user_prompt += "Label: " + str(current_label) + "\n\n"

            user_prompt += f"Question: {query}\nDocument: {document}\nAnswer: {answer}\nLabel: "

            time.sleep(request_delay)

            if debug_mode: 
                print("------------------------------------------")
                print("Answer Relevance Evaluation")
                print(user_prompt)

            client = anthropic.Anthropic(
            api_key=ANTHROPIC_API_KEY,
            )

            responses = []

            with client.messages.stream( 
                max_tokens=1024, 
                system=system_prompt.strip(),
                messages=[{"role": "user", "content": user_prompt.strip()}], 
                model=model_choice
            ) as stream: 
                for text in stream.text_stream: 
                    responses.append(text)

            final_response = " ".join(responses)

            if debug_mode: 
                print(final_response)

            yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
            no = r"\[\s*\[?\s*No\s*\]?\s*\]"
            
            if re.search(yes, final_response): 
                if debug_mode is True:
                    print("Returned label 1")
                return 1
            elif re.search(no, final_response):
                if debug_mode is True:
                    print("Returned label 0")
                return 0
            else:
                print("Didn't extract Yes or No!")
                failed_extraction_count['failed'] += 1
                return 0
        except Exception as e:
            if attempt < 4:  # Only print the error message if not on the last attempt
                print(f"Attempt {attempt + 1} failed with error: {e}")
                time.sleep(60)  # Sleep for 60 seconds before retrying
            else:
                print("All attempts failed. Last error was:", e)
                return 0
            
# Define the no-op functions
def no_op_function(*args, **kwargs):
    print("vLLM is not available. This function call is ignored.")

if VLLM_AVAILABLE:
    def few_shot_context_relevance_scoring_vllm(
        system_prompt: str, query: str, document: str, model_choice: str, 
        query_id: str, debug_mode: bool, host_url: str, request_delay: int, 
        failed_extraction_count: Dict[str,int] = {'failed': 0}, 
        few_shot_examples=None
    ) -> int:
        """
        Evaluates the relevance of a query given a document using few-shot examples.
        This function constructs a user prompt with few-shot examples (if provided) and the current query,
        document, and answer. It then queries a local vLLM model to determine if the query is context relevant to the
        document.

        Args:
        system_prompt (str): The prompt for the system role in the conversation.
        query (str): The query to be evaluated.
        document (str): The document to be evaluated.
        model_choice (str): The model identifier for vLLM's model.
        query_id (str): The identifier for the query column in few_shot_examples.
        debug_mode (bool): Flag to turn on debug mode for additional output.
        host_url (str): The URL is used to send requests to the locally running model.
        request_delay (int): Time in seconds to delay the request (simulating network latency).
        failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
        few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

        Returns:
        int: The label indicating context relevance (1 for relevant, 0 for not relevant).

        Raises:
        Warning: If incorrect labels are detected in the few-shot examples.
        """
        for attempt in range(5):
            try:
                user_prompt = ""
                if few_shot_examples is not None:
                    for row in range(len(few_shot_examples)):
                        user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                        user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                        current_label = few_shot_examples.iloc[row]['Context_Relevance_Label']
                        if current_label in {1, 1.0}:
                            warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                            current_label = "[[Yes]]"
                        elif current_label in {0, 0.0}:
                            warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                            current_label = "[[No]]"
                        user_prompt += "Label: " + str(current_label) + "\n\n"
                
                user_prompt += f"Question: {query}\nDocument: {document}\nLabel: "

                time.sleep(request_delay)

                if debug_mode: 
                    print("------------------------------------------")
                    print("Context Relevance Evaluation:")
                    print(user_prompt)
                
                openai_api_key = "EMPTY"
                client = OpenAI(
                api_key=openai_api_key,
                base_url=host_url
                )

                chat_completion = client.chat.completions.create(
                messages=[
                    {
                    "role": "system",
                    "content": system_prompt,
                    },
                    {
                    "role": "user",
                    "content": user_prompt,
                    }
                ],
                model=model_choice
                )
                final_response = chat_completion.choices[0].message.content

                if debug_mode: 
                    print(final_response)

                yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
                no = r"\[\s*\[?\s*No\s*\]?\s*\]"

                if re.search(yes, final_response): 
                    if debug_mode:
                        print("Returned label 1")
                    return 1
                elif re.search(no, final_response):
                    if debug_mode:
                        print("Returned label 0")
                    return 0
                else:
                    print("Didn't extract Yes or No!")
                    failed_extraction_count['failed'] += 1
                    return 0
            except Exception as e:
                if attempt < 4:  # Only print the error message if not on the last attempt
                    print(f"Attempt {attempt + 1} failed with error: {e}")
                    time.sleep(60)  # Sleep for 60 seconds before retrying
                else:
                    print("All attempts failed. Last error was:", e)
                    return 0
        
    def few_shot_answer_faithfulness_scoring_vllm(
        system_prompt: str, query: str, document: str, answer: str, model_choice: str, 
        query_id: str, debug_mode: bool, host_url: str, request_delay: int, 
        failed_extraction_count: Dict[str,int] = {'failed': 0}, 
        few_shot_examples=None
    ) -> int:
        """
        Evaluates the faithfulness of an answer given a document and a query using few-shot examples.
        This function constructs a user prompt with few-shot examples (if provided) and the current query,
        document, and answer. It then queries a local vLLM model to determine if the answer is faithful.

        Args:
        system_prompt (str): The prompt for the system role in the conversation.
        query (str): The query related to the document.
        document (str): The document to be evaluated.
        answer (str): The answer to be evaluated for faithfulness.
        model_choice (str): The model identifier for vLLM's model.
        query_id (str): The identifier for the query column in few_shot_examples.
        debug_mode (bool): Flag to turn on debug mode for additional output.
        host_url (str): The URL is used to send requests to the locally running model.
        request_delay (int): Time in seconds to delay the request (simulating network latency).
        failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
        few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

        Returns:
        int: The label indicating answer faithfulness 1 if the answer is faithful, 0 otherwise.

        Raises:
        Warning: If incorrect labels are detected in the few-shot examples.
        """
        for attempt in range(5):
            try:
                user_prompt = ""
                if few_shot_examples is not None:
                    for row in range(len(few_shot_examples)):
                        user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                        user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                        user_prompt += f"Answer: {few_shot_examples.iloc[row]['Answer']}\n"
                        current_label = few_shot_examples.iloc[row]['Answer_Faithfulness_Label']
                        if current_label in {1, 1.0}:
                            warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                            current_label = "[[Yes]]"
                        elif current_label in {0, 0.0}:
                            warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                            current_label = "[[No]]"
                        user_prompt += "Label: " + str(current_label) + "\n\n"

                user_prompt += f"Question: {query}\nDocument: {document}\nAnswer: {answer}\nLabel: "

                time.sleep(request_delay) # Model query delay chosen by user

                if debug_mode: 
                    print("------------------------------------------")
                    print("Answer Faithfulness Evaluation")
                    print(user_prompt)

                openai_api_key = "EMPTY"
                client = OpenAI(
                api_key=openai_api_key,
                base_url=host_url
                )

                chat_completion = client.chat.completions.create(
                messages=[
                    {
                    "role": "system",
                    "content": system_prompt,
                    },
                    {
                    "role": "user",
                    "content": user_prompt,
                    }
                ],
                model=model_choice
                )
                final_response = chat_completion.choices[0].message.content
                
                if debug_mode: 
                    print(final_response)

                yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
                no = r"\[\s*\[?\s*No\s*\]?\s*\]"
                
                if re.search(yes, final_response): 
                    if debug_mode:
                        print("Returned label 1")
                    return 1
                elif re.search(no, final_response):
                    if debug_mode:
                        print("Returned label 0")
                    return 0
                else:
                    print("Didn't extract Yes or No!")
                    failed_extraction_count['failed'] += 1
                    return 0
            except Exception as e:
                if attempt < 4:  # Only print the error message if not on the last attempt
                    print(f"Attempt {attempt + 1} failed with error: {e}")
                    time.sleep(60)  # Sleep for 60 seconds before retrying
                else:
                    print("All attempts failed. Last error was:", e)
                    return 0
            
    def few_shot_answer_relevance_scoring_vllm(
        system_prompt: str, query: str, document: str, answer: str, model_choice: str, 
        query_id: str, debug_mode: bool, host_url: str, request_delay: int, 
        failed_extraction_count: Dict[str,int] = {'failed': 0}, 
        few_shot_examples=None
    ) -> int:
        """
        Evaluates the relevance of an answer given a document and a query using few-shot examples.
        This function constructs a user prompt with few-shot examples (if provided) and the current query,
        document, and answer. It then queries a local vLLM model to determine if the answer is relevant.

        Args:
        system_prompt (str): The prompt for the system role in the conversation.
        query (str): The query related to the document.
        document (str): The document to be evaluated.
        answer (str): The answer to be evaluated for relevance.
        model_choice (str): The model identifier for vLLM's model.
        query_id (str): The identifier for the query column in few_shot_examples.
        debug_mode (bool): Flag to turn on debug mode for additional output.
        host_url (str): The URL is used to send requests to the locally running model.
        request_delay (int): Time in seconds to delay the request (simulating network latency).
        failed_extraction_count (Dict[str, int]): A dictionary to count the number of failed extractions.
        few_shot_examples (DataFrame, optional): A DataFrame containing few-shot examples.

        Returns:
        int: The label indicating answer relevance 1 if the answer is relevant, 0 otherwise.

        Raises:
        Warning: If incorrect labels are detected in the few-shot examples.
        """
        for attempt in range(5):
            try:
                user_prompt = ""
                if few_shot_examples is not None:
                    for row in range(len(few_shot_examples)):
                        user_prompt += f"Question: {few_shot_examples.iloc[row][query_id]}\n"
                        user_prompt += f"Document: {few_shot_examples.iloc[row]['Document']}\n"
                        user_prompt += f"Answer: {few_shot_examples.iloc[row]['Answer']}\n"
                        current_label = few_shot_examples.iloc[row]['Answer_Relevance_Label']
                        if current_label in {1, 1.0}:
                            warnings.warn("Incorrect label '1' detected. Please use '[[Yes]]' instead.")
                            current_label = "[[Yes]]"
                        elif current_label in {0, 0.0}:
                            warnings.warn("Incorrect label '0' detected. Please use '[[No]]' instead.")
                            current_label = "[[No]]"
                        user_prompt += "Label: " + str(current_label) + "\n\n"
                
                user_prompt += f"Question: {query}\nDocument: {document}\nAnswer: {answer}\nLabel: "

                time.sleep(request_delay) # Model query delay chosen by user

                if debug_mode: 
                    print("------------------------------------------")
                    print("Answer Relevance Evaluation")
                    print(user_prompt)

                openai_api_key = "EMPTY"
                client = OpenAI(
                api_key=openai_api_key,
                base_url=host_url
                )

                chat_completion = client.chat.completions.create(
                messages=[
                    {
                    "role": "system",
                    "content": system_prompt,
                    },
                    {
                    "role": "user",
                    "content": user_prompt,
                    }
                ],
                model=model_choice
                )
                final_response = chat_completion.choices[0].message.content
                
                if debug_mode is True: 
                    print(final_response)

                yes = r"\[\s*\[?\s*Yes\s*\]?\s*\]"
                no = r"\[\s*\[?\s*No\s*\]?\s*\]"
                
                if re.search(yes, final_response): 
                    if debug_mode is True:
                        print("Returned label 1")
                    return 1
                elif re.search(no, final_response):
                    if debug_mode is True:
                        print("Returned label 0")
                    return 0
                else:
                    print("Didn't extract Yes or No!")
                    failed_extraction_count['failed'] += 1
                    return 0
            except Exception as e:
                if attempt < 4:  # Only print the error message if not on the last attempt
                    print(f"Attempt {attempt + 1} failed with error: {e}")
                    time.sleep(60)  # Sleep for 60 seconds before retrying
                else:
                    print("All attempts failed. Last error was:", e)
                    return 0
else:
    # Override the functions with the no-op function if vLLM is not available
    few_shot_context_relevance_scoring_vllm = no_op_function
    few_shot_answer_faithfulness_scoring_vllm = no_op_function
    few_shot_answer_relevance_scoring_vllm = no_op_function


context_relevance_system_prompt = (
    "You are an expert dialogue agent. "
    "Your task is to analyze the provided document and determine whether it is relevant for responding to the dialogue. "
    "In your evaluation, you should consider the content of the document and how it relates to the provided dialogue. "
    'Output your final verdict by strictly following this format: "[[Yes]]" if the document is relevant and "[[No]]" if the document provided is not relevant. '
    "Do not provide any additional explanation for your decision.\n\n"
)

answer_relevance_system_prompt = (
    "Given the following question, document, and answer, you must analyze the provided answer and document before determining whether the answer is relevant for the provided question. "
    "In your evaluation, you should consider whether the answer addresses all aspects of the question and provides only correct information from the document for answering the question. "
    'Output your final verdict by strictly following this format: "[[Yes]]" if the answer is relevant for the given question and "[[No]]" if the answer is not relevant for the given question. '
    "Do not provide any additional explanation for your decision.\n\n"
)

answer_faithfulness_system_prompt = (
    "Given the following question, document, and answer, you must analyze the provided answer and determine whether it is faithful to the contents of the document. "
    "The answer must not offer new information beyond the context provided in the document. "
    "The answer also must not contradict information provided in the document. "
    'Output your final verdict by strictly following this format: "[[Yes]]" if the answer is faithful to the document and "[[No]]" if the answer is not faithful to the document. '
    "Do not provide any additional explanation for your decision.\n\n"
)


from .LLM_as_a_Judge_Adaptation.General_Binary_Classifier import (
    load_model,
    prepare_and_clean_data,
    analyze_and_report_data,
    transform_data,
    split_dataset,
    prepare_dataset,
    initalize_dataset_for_tokenization,
    train_and_evaluate_model,
    evaluate_model,
    print_and_save_model,
    set_random_seed
)
from tqdm import tqdm
import torch
import pandas as pd
import sys
import warnings

def binary_classifer_config(
    training_dataset: list,
    validation_set: list,
    label_column: list,
    num_epochs: int,
    patience_value: int,
    learning_rate: float,
    training_dataset_path: str = "None",
    validation_dataset_path: str = "None",
    model_choice: str = "microsoft/deberta-v3-large",
    validation_set_scoring: bool = True,
    assigned_batch_size: int = 1,
    gradient_accumulation_multiplier: int = 32,
    number_of_runs: int = 1,
    num_warmup_steps: int = 100,
    training_row_limit: int = -1,
    validation_row_limit: int = -1
) -> None:
    """
    Configures and runs the binary classifier training and evaluation pipeline.

    Parameters:
    - training_dataset (list): List of paths to training datasets.
    - validation_set (list): List of paths to validation datasets.
    - label_column (list): List of label columns.
    - num_epochs (int): Number of epochs for training.
    - patience_value (int): Patience value for early stopping.
    - learning_rate (float): Learning rate for the optimizer.
    - training_dataset_path (str): Path to save the combined training dataset.
    - validation_dataset_path (str): Path to save the combined validation dataset.
    - model_choice (str): Model choice for the classifier.
    - validation_set_scoring (bool): Whether to score the validation set.
    - assigned_batch_size (int): Batch size for training.
    - gradient_accumulation_multiplier (int): Gradient accumulation multiplier.
    - number_of_runs (int): Number of runs for training.
    - num_warmup_steps (int): Number of warmup steps for the optimizer.
    - training_row_limit (int): Limit on the number of rows for training datasets.
    - validation_row_limit (int): Limit on the number of rows for validation datasets.

    Returns:
    - None
    """

    device = torch.device("cuda:0")
    learning_rate_choices = [learning_rate]

    if training_row_limit == -1:
        training_row_limit = None

    if validation_row_limit == -1:
        validation_row_limit = None

    if len(training_dataset) == 1:
        training_dataset_path = training_dataset[0]
    elif training_dataset_path == "None":
        sys.exit("Error: A path for saving the combined dataset is required when multiple training datasets are specified. Please specify the 'training_dataset_path' parameter in your configuration. This path will be used to store the dataset resulting from merging all provided training datasets.")
    else:
        combined_df = pd.DataFrame()
        print("\n")
        for dataset in tqdm(training_dataset, desc="Combining datasets"):
            temp_df = pd.read_csv(dataset, sep='\t', low_memory=False, nrows=training_row_limit)
            print(f"Loaded dataset {dataset} with {temp_df.shape[0]} rows" + 
                  (f" (limited to first {training_row_limit} rows)" if training_row_limit is not None else ""))
            combined_df = pd.concat([combined_df, temp_df], ignore_index=True)
        combined_df.to_csv(training_dataset_path, sep='\t', index=False)

    if len(validation_set) == 1:
        validation_dataset_path = validation_set[0]
    elif len(validation_set) > 1:
        if validation_dataset_path == "None":
            sys.exit("Error: A path for saving the combined validation dataset is required when multiple validation sets are specified. Please specify the 'validation_dataset_path'.")
        elif len(validation_set) != len(training_dataset):
            warnings.warn(f"Error: Validations sets provided are not equivalent to number of training datasets. {len(validation_set)} validation set(s) are provided, and {len(training_dataset)} are provided")
        else:
            combined_val_df = pd.DataFrame()
            print("--------------------------------------------------------------------------")
            for dataset in tqdm(validation_set, desc="Combining validation datasets"):
                temp_val_df = pd.read_csv(dataset, sep='\t', low_memory=False, nrows=validation_row_limit)
                print(f"Loaded dataset {dataset} with {temp_val_df.shape[0]} rows" + 
                      (f" (limited to first {validation_row_limit} rows)" if validation_row_limit is not None else ""))
                combined_val_df = pd.concat([combined_val_df, temp_val_df], ignore_index=True)
            combined_val_df.to_csv(validation_dataset_path, sep='\t', index=False)
            print("--------------------------------------------------------------------------")

            initial_count = combined_val_df.shape[0]
            combined_val_df = combined_val_df.dropna(subset=['Answer'])
            final_count = combined_val_df.shape[0]

            removed_count = initial_count - final_count
            if removed_count > 0:
                print(f"Removed {removed_count} data points from the validation set due to NaN in the 'Answer' column.")

            combined_val_df.to_csv(validation_dataset_path, sep='\t', index=False)

    for label in label_column:
        for chosen_learning_rate in learning_rate_choices:
            set_random_seed(42)

            tokenizer, max_token_length = load_model(model_choice)

            prepare_data_settings = {
                "training_dataset_path": training_dataset_path,
                "learning_rate_choices": learning_rate_choices,
                "chosen_learning_rate": chosen_learning_rate,
                "model_choice": model_choice,
                "number_of_runs": number_of_runs,
                "validation_set_scoring": validation_set_scoring,
                "label": label,
                "validation_dataset_path": validation_dataset_path,
                "patience_value": patience_value,
                "num_epochs": num_epochs,
                "num_warmup_steps": num_warmup_steps,
                "gradient_accumulation_multiplier": gradient_accumulation_multiplier,
                "assigned_batch_size": assigned_batch_size,
                "tokenizer": tokenizer
            }

            checkpoint_path, patience_value = prepare_and_clean_data(prepare_data_settings)

            synth_queries = analyze_and_report_data(training_dataset_path, label, tokenizer, max_token_length)

            train_df, test_set = transform_data(synth_queries, validation_dataset_path, label)

            train_set_text, train_set_label, dev_set_text, dev_set_label, test_set_text, text_set_label_, labels_list = split_dataset(train_df, training_dataset_path, test_set, label)

            training_dataset_pandas, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow, test_dataset_pandas = prepare_dataset(validation_set_scoring, train_set_label, train_set_text, dev_set_label, dev_set_text)

            tokenized_datasets = initalize_dataset_for_tokenization(tokenizer, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow)

            train_and_eval_settings = {
                "number_of_runs": number_of_runs,
                "tokenized_datasets": tokenized_datasets,
                "assigned_batch_size": assigned_batch_size,
                "train_set_label": train_set_label,
                "model_choice": model_choice,
                "chosen_learning_rate": chosen_learning_rate,
                "device": device,
                "checkpoint_path": checkpoint_path,
                "patience_value": patience_value,
                "num_epochs": num_epochs,
                "num_warmup_steps": num_warmup_steps,
                "gradient_accumulation_multiplier": gradient_accumulation_multiplier
            }

            model, avg_train_losses, avg_valid_losses, eval_dataloader, inference_times = train_and_evaluate_model(train_and_eval_settings)

            total_predictions, total_references, metric = evaluate_model(model, model_choice, checkpoint_path, device, eval_dataloader, inference_times)

            print_and_save_model(total_predictions, total_references, checkpoint_path, metric)
